{# query_result.html #}
{% extends "base.html" %}

{% comment %}
    template parameters:
        sql_highlight_css: str
        table_html: str

    ... and as per
        query_nav.html
        page_nav_results.html
{% endcomment %}

{% block extrahead %}
    <style type="text/css">{{ sql_highlight_css }}</style>
{% endblock %}

{% block content %}

    {% include "query_nav.html" %}

    <h1>Results</h1>

    {% include "page_nav_results.html" with page=page %}

    {% include "expand_collapse_buttons.html" %}

    {{ table_html|safe }}

    {% include "page_nav_results.html" with page=page %}

    <h1>SQL</h1>
    {{ sql|safe }}

{% endblock %}
