{# global_header.html #}

<div>
    <i>CRATE: Clinical Records Anonymisation and Text Extraction</i>
</div>
<div>
    Database: <b>{{ RESEARCH_DB_TITLE }}</b>.
    User: <b>{{ user }}</b>.
    [
    {% if nav_on_main_menu %}
        Main menu
    {% else %}
        <a href="{% url 'home' %}">Main menu</a>
    {% endif %}
    | <a href="{% url 'logout' %}">Logout</a>
    ]
</div>
