{# admin/readonly_view_form.html #}

{% extends "admin/change_form.html" %}
{% load i18n admin_urls admin_static admin_modify %}

{% block submit_buttons_top %}{% endblock %}
{% block submit_buttons_bottom %}{% endblock %}
  {# ... REMOVED: {% submit_row %} #}
