{# edit_profile.html #}

{% extends "base.html" %}

{% block content %}

    <h1>Edit personal settings</h1>

    <form action="{% url 'edit_profile' %}" method="post">
        {% csrf_token %}
        <table class="formtable">
            {{ form.as_table }}
        </table>
        <input type="submit" value="Submit" />
    </form>

{% endblock %}
